#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import os

from config import Config
from api import API
from group import Group
from user import User
from host import Host
from layer import Layer
from resource import Resource
from role import Role
from variable import Variable
from exceptions import ConjurException

from config import config

def _config(given):
    if given is None:
        env_config_file = os.getenv('CONJURRC')
        if env_config_file is not None:
            config.load(env_config_file)
        return config
    return given


def configure(**kwargs):
    """
    Convenience function to merge multiple settings into the default global
    config.

    Example:

        >>> import conjur
        >>> conjur.configure(appliance_url='https://conjur.example.com/api',
        ...              account='example',
        ...              cert_path='/path/to/cert.pem')

    """
    config.update(**kwargs)
    return config


def new_from_netrc(netrc_file=None, configuration=None):
    """
    Create a `conjur.API` instance using an identity loaded from netrc.  This method
    uses the identity stored for the host `config.authn_url`.

    `netrc_file` is an alternative path to the netrc formatted file.  Defaults
    to ~/.netrc on unixy systems.

    `configuration` is a `conjur.Config` instance used to determine the host
    in the netrc file, and also passed to the `conjur.new_from_key` method to
    create the API instance using the identity.
    """
    import netrc

    configuration = _config(configuration)
    auth = netrc.netrc(netrc_file).authenticators(configuration.authn_url)
    if auth is None:
        raise ValueError("No authenticators found for authn_url '%s' in %s" % (
            configuration.authn_url,
            (netrc_file or '~/.netrc')
        ))
    login, _, api_key = auth
    return new_from_key(login, api_key, configuration)


def new_from_key(login, api_key, configuration=None):
    """
    Create a `conjur.API` instance that will authenticate on demand as the identity given
    by `login` and `api_key`.

    `login` is the identity of the Conjur user or host to authenticate as.

    `api_key` is the api key *or* password to use when authenticating.

    `configuration` is a `conjur.Config` instance for the api.  If not given the global
        `Config` instance (`conjur.config`) will be used.
    """

    return API(credentials=(login, api_key), config=_config(configuration))


def new_from_token(token, configuration=None):
    """
    Create a `conjur.API` instance that will authenticate using the given signed Conjur
    token.

    This is useful if you want to act on behalf of a the identity of an
    HTTP request containing a user's signed token.

    `token` is the json formatted, *not* base64'd, Conjur authentication Token.

    `configuration` is a conjur.Config instance for the api.  If not given, the global Config
    instance (`conjur.config`) will be used.
    """
    return API(token=token, config=_config(configuration))

__all__ = (
    'config', 'Config', 'Group', 'API', 'User', 'Host', 'Layer', 'Resource', 'Role', 'Variable',
    'new_from_key', 'new_from_netrc', 'new_from_token', 'configure', 'ConjurException'
)


#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


class Group(object):
    """
    Represents a Conjur [group](https://developer.conjur.net/reference/services/directory/group).

    Generally you won't create instances of this class, but use the `conjur.API.group(id)` method.

    A group is a role that contains other roles, typically users and other groups.  A `conjur.Group`
    object can list its members with the `members` method, and also manage them with the `add_member`
    and `remove_member` methods.
    """
    def __init__(self, api, id):
        self.api = api
        """
        Instance of `conjur.API` used to implement Conjur operations.
        """

        self.id = id
        """
        Identifier (unqualified) of the group.
        """

        self.role = api.role('group', id)
        """
        Represents the `conjur.Role` associated with this group.
        """

    def members(self):
        """
        Return a list of members of this group.  Members are returned as `dict`s
        with the following keys:

        * `'member'` the fully qualified identifier of the group
        * `'role'` the fully qualified identifier of the group (redundant)
        * `'grantor'` the role that granted the membership
        * `'admin_option'` whether this member can grant membership in the group to other roles.

        Example: print member ids (fully qualified) and whether they are admins of the group.
            >>> group = api.group('security_admin')
            >>> for member in group.members():
            ...     print('{} is a member of security_admin ({} admin option)'.format(
            ...         member['member'],
            ...         'with' if member['admin_option'] else 'without'
            ...     ))
        """
        return self.role.members()

    def add_member(self, member, admin=False):
        """
        Add a member to this group.

        `member` is the member we want to add to the group, and  should be a qualified Conjur id,
            or an object with a `role` attribute or a `roleid` method.  Examples of such objects
            include `conjur.User`, `conjur.Role`, and `conjur.Group`.

        If `admin` is True, the member will be allowed to add other members to this group.
        """
        self.role.grant_to(member, admin)

    def remove_member(self, member):
        """
        Remove a member from the group.

        `member` is the member to remove, and  should be a qualified Conjur id,
            or an object with a `role` attribute or a `roleid` method.  Examples of such objects
            include `conjur.User`, `conjur.Role`, and `conjur.Group`.
        """
        self.role.revoke_from(member)



#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import base64

from mock import patch, Mock
import requests

import conjur


@patch.object(requests, 'post')
def test_authenticate(mock_post):
    api = conjur.new_from_key("login", "api-key")
    api.config.authn_url = "https://example.com"
    mock_post.return_value = mock_response = Mock()
    mock_response.status_code = 200
    mock_response.text = "token token token"
    token = api.authenticate()
    assert token == "token token token"
    mock_post.assert_called_with("https://example.com/users/login/authenticate",
                                 "api-key", verify=api.config.verify)


@patch.object(requests, 'post')
def test_authenticate_with_cached_token(mock_post):
    api = conjur.new_from_token("token token")
    assert api.authenticate() == "token token"
    mock_post.assert_not_called()


def test_auth_header():
    api = conjur.new_from_token("the token")
    expected = 'Token token="%s"' % (base64.b64encode("the token"))
    assert api.auth_header() == expected


@patch.object(requests, 'post')
def test_request_with_post(mock_post):
    api = conjur.new_from_token("the token")
    mock_post.return_value = resp = Mock()
    resp.status_code = 200
    assert api.request('post', 'https://example.com', data="foobar") == resp
    mock_post.assert_called_with('https://example.com', data="foobar",
                                 headers={"Authorization": api.auth_header()},
                                 verify=api.config.verify)



#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from mock import patch, Mock
from conjur import Config, new_from_key
config = Config()
api = new_from_key('login', 'pass', config)
config.account = 'the-account'
config.appliance_url = 'https://example.com/api'


def test_roleid():
    role = api.role('some-kind', 'the-id')
    assert role.roleid == 'the-account:some-kind:the-id'


@patch.object(api, 'put')
def test_role_grant_to_without_admin(mock_put):
    role = api.role('some-kind', 'the-id')
    role.grant_to('some-other-role')
    mock_put.assert_called_with(
        '{0}/the-account/roles/some-kind/the-id?members&member={1}'.format(
            config.authz_url,
            'some-other-role'
        ),
        data={}
    )


@patch.object(api, 'put')
def test_role_grant_to_with_admin_true(mock_put):
    role = api.role('some-kind', 'the-id')
    role.grant_to('some-other-role', True)
    mock_put.assert_called_with(
        '{0}/the-account/roles/some-kind/the-id?members&member={1}'.format(
            config.authz_url,
            'some-other-role'
        ),
        data={'admin': 'true'}
    )


@patch.object(api, 'put')
def test_role_grant_to_with_admin_false(mock_put):
    role = api.role('some-kind', 'the-id')
    role.grant_to('some-other-role', False)
    mock_put.assert_called_with(
        '{0}/the-account/roles/some-kind/the-id?members&member={1}'.format(
            config.authz_url,
            'some-other-role'
        ),
        data={'admin': 'false'}
    )


@patch.object(api, 'delete')
def test_role_revoke_from(mock_del):
    role = api.role('some-kind', 'the-id')
    role.revoke_from('some-other-role')
    mock_del.assert_called_with(
        '{0}/the-account/roles/some-kind/the-id?members&member={1}'.format(
            config.authz_url,
            'some-other-role'
        )
    )


@patch.object(api, 'get')
def test_role_members(mock_get):
    members = ['foo', 'bar']
    mock_get.return_value = Mock(json=lambda: members)
    role = api.role('blah', 'boo')
    assert role.members() == members
    mock_get.assert_called_with(
        '{0}/the-account/roles/blah/boo?members'.format(config.authz_url)
    )


@patch.object(api, 'put')
def test_role_grant_to_user(mock_put):
    role = api.role('somekind', 'admins')
    user = api.user('somebody')
    role.grant_to(user)
    mock_put.assert_called_with(
        '{0}/the-account/roles/somekind/admins?members&member={1}'.format(
            config.authz_url,
            'the-account%3Auser%3Asomebody'
        ),
        data={}
    )



#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal inre
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from conjur.util import authzid
from conjur.role import Role
from conjur.exceptions import ConjurException


class Resource(object):
    """
    A `Resource` represents an object on which `Role`s can
    be permitted to perform certain actions.

    Generally you will not construct these directly, but call the `conjur.API.role` method
    to do so.
    """
    def __init__(self, api, kind, identifier):
        self.api = api
        self.kind = kind
        self.identifier = identifier

    @property
    def resourceid(self):
        """
        The fully qualified resource id as a string, like `'the-account:variable:db-password`.
        """
        return ":".join([self.api.config.account, self.kind, self.identifier])

    def permit(self, role, privilege, grant_option=False):
        """
        Permit `role` to perform `privilege` on this resource.

        `role` is a qualified conjur identifier (e.g. `'user:alice`') or an object
            with a `role` attribute or `roleid` method, such as a `conjur.User` or
            `conjur.Group`.

        If `grant_option` is True, the role will be able to grant this
        permission to other resources.

        You must own the resource or have the permission with `grant_option`
        to call this method.
        """
        data = {}
        params = {
            'permit': 'true',
            'privilege': privilege,
            'role': authzid(role, 'role')
        }
        if grant_option:
            data['grant_option'] = 'true'

        self.api.post(self.url(), data=data, params=params)

    def deny(self, role, privilege):
        """
        Deny `role` permission to perform `privilege` on this resource.

        You must own the resource or have the permission with `grant_option`
        on it to call this method.
        """
        params = {
            'permit': 'true',
            'privilege': privilege,
            'role': authzid(role)
        }

        self.api.post(self.url(), params=params)

    def permitted(self, privilege, role=None):
        """
        Return True if `role` has `privilege` on this resource.

        `role` may be a `conjur.Role` instance, an object with a `role` method,
        or a qualified role id as a string.

        If `role` is not given, check the permission for the currently
        authenticated role.

        Example: Check that the currently authenticated role is allowed to `'read'`
            a resource.

            >>> service = api.resource('service', 'gateway')
            >>> if service.permitted('read'):
            ...     print("I can read 'service:gateway'")
            ... else:
            ...     print("I cannot read 'service:gateway'")

        Example: Check whether members of group 'security_admin' can 'update'
            a resource.

            >>> service = api.resource('service', 'gateway')
            >>> security_admin = api.group('security_admin')
            >>> if service.permitted('update', security_admin):
            ...     print("security_admin members can update service:gateway")
            >>> else:
            ...     print("security_admin members cannot update service:gateway")
        """

        if role is None:
            # Handle self role check
            response = self.api.get(self.url(),
                                    params={'check': 'true',
                                            'privilege': privilege},
                                    check_errors=False)
            if response.status_code == 204:
                return True
            elif response.status_code in (404, 403):
                return False
            else:
                raise ConjurException("Request failed: %d" % response.status_code)
        else:
            # Otherwise call role.is_permitted
            return Role.from_roleid(self.api, role).is_permitted(self, privilege)

    def url(self):
        """
        Internal method to return a url for this object as a string.
        """
        return "/".join([
            self.api.config.authz_url,
            self.api.config.account,
            'resources',
            self.kind,
            self.identifier
        ])



#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Defines configuration settings needed to connect to a Conjur appliance.

For convenience, a "global" `conjur.config.Config` instance is available as
`conjur.config.config`.  This object is used when a `conjur.api.API` instance
is created without a config.

Example:
    ```python
    from conjur.config import config

    # Set multiple keys using the `update

The most important settings are `appliance_url`, `account`, and `cert_file`.

  * `appliance_url` points the client to your Conjur appliance.  Suppose the hostname of
    the appliance is `conjur.example.com`.  Then the appliance url is `https://conjur.example.com/api`.
    Note that the `scheme` **must** be `https`, and the `/api` path is required.

  * `account` is an organizational name chosen when you deploy your conjur appliance.  The CLI command
     `conjur authn whoami` will show you this value.  This value is required.

  * `cert_file` is the path to a `pem` formated certificate used to make a secure connection to the
     Conjur appliance.  The CLI command `conjur init -h <conjur hostname>` is the standard way to install
     such a certificate.  This option is **required** if you are using a self-signed certificate (which is
     the most common use case).

While endpoints for specific services can also be configured, this is not normally needed
outside of development work.
"""

import os

_DEFAULT = object()

class ConfigException(Exception):
    pass


def _setting(name, default=_DEFAULT, doc=''):
    def fget(self):
        return self.get(name, default)

    def fset(self, value):
        self.set(name, value)

    return property(fget, fset, doc=doc)


def _service_url(name, per_account=True, doc=''):
    def fget(self):
        return self.service_url(name, per_account)

    def fset(self, value):
        self.set(name + '_url', value)

    return property(fget=fget, fset=fset, doc=doc)


class Config(object):
    def __init__(self, **kwargs):
        self._config = {}
        self.update(kwargs)

    def load(self, input):
        import yaml

        if isinstance(input, str):
            input = open(input, 'r')
        conf = yaml.safe_load(input)
        self.update(conf)

    def update(self, *dicts, **kwargs):
        for d in dicts + (kwargs, ):
            self._config.update(d)

    def service_url(self, service, per_account=True):
        key = '%s_url' % service
        if key in self._config:
            return self._config[key]
        if self.appliance_url is not None:
            url_parts = [self.appliance_url]
            if service != "core":
                url_parts.append(service)
            return "/".join(url_parts)
        else:
            raise ConfigException('Missing appliance_url')

    def get(self, key, default=_DEFAULT):
        if key in self._config:
            return self._config[key]
        env_key = 'CONJUR_' + key.upper()
        if env_key in os.environ:
            value = os.environ[env_key]
            self._config[key] = value
            return value
        if default is _DEFAULT:
            raise Exception("config setting %s is required" % key)
        return default

    def set(self, key, value):
        self._config[key] = value

    authn_url = _service_url('authn', doc='URL for the authn service')
    core_url = _service_url('core', doc='URL for the core service')
    authz_url = _service_url('authz',
                             per_account=False,
                             doc='URL for the authz service')

    pubkeys_url = _service_url('pubkeys', doc='URL for the pubkeys service')


    cert_file = _setting('cert_file', None,
                         "Path to certificate to verify ssl requests \
                         to appliance")

    account = _setting('account', 'conjur', 'Conjur account identifier')

    appliance_url = _setting('appliance_url', None, 'URL for Conjur appliance')

    @property
    def verify(self):
        '''
        Argument to be passed to `requests` methods `verify` keyword argument.
        '''
        if self.cert_file is not None:
            return self.cert_file
        else:
            return True


config = Config()
"""
A global `conjur.Config` instance, used when creating an `conjur.API` instance
by default.
"""

default = config

__all__ = ('Config', 'config', 'default')



#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from conjur.util import urlescape


class User(object):
    def __init__(self, api, login, attrs=None):
        self.api = api
        self.login = login
        # support as_role
        self.role = api.role('user', login)
        self._attrs = attrs

    def exists(self):
        resp = self.api.get(self.url(), check_errors=False)
        return resp.status_code != 404

    def __getattr__(self, item):
        if self._attrs is None:
            self._fetch()
        try:
            return self._attrs[item]
        except KeyError:
            raise AttributeError(item)

    def _fetch(self):
        self._attrs = self.api.get(self.url()).json()

    def url(self):
        return "{0}/users/{1}".format(self.api.config.core_url,
                                      urlescape(self.login))



#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from mock import patch
import conjur

api = conjur.new_from_key('foo', 'bar')

group = api.group('v1/admins')


def test_group():
    assert group.role.kind == 'group'
    assert group.role.identifier == 'v1/admins'
    assert group.role.roleid == api.config.account + ':group:v1/admins'


@patch.object(group.role, 'grant_to')
def test_add_member(mock_grant_to):
    member = api.user('foo')
    group.add_member(member)
    mock_grant_to.assert_called_with(member, False)


@patch.object(group.role, 'grant_to')
def test_add_member_admin(mock_grant_to):
    member = api.role('something', 'else')
    group.add_member(member, True)
    mock_grant_to.assert_called_with(member, True)


@patch.object(group.role, 'revoke_from')
def test_remove_member(mock_revoke_from):
    member = api.user('foo')
    group.remove_member(member)
    mock_revoke_from.assert_called_with(member)



#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import conjur
from conjur.config import Config, config


def test_new_from_key():
    api = conjur.new_from_key("login", "secret")
    assert api.token is None
    assert api.api_key == "secret"
    assert api.login == "login"
    assert api.config == config


def test_new_from_token():
    api = conjur.new_from_token("a token")
    assert api.token == "a token"
    assert api.api_key is None
    assert api.login is None
    assert api.config == config


def test_new_with_config():
    cfg = Config()
    api = conjur.new_from_key("login", "secret", cfg)
    assert cfg is api.config



#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from conjur.util import urlescape
from conjur.exceptions import ConjurException


class Host(object):
    """
    A Conjur `Host` is a role corresponding to a machine or machine identity.

    The `Host` class provides the ability to check for existence and read attributes of the
    host.

    Attributes (such as the `ownerid`) are fetched lazily.

    Newly created hosts, as returned by `conjur.API.create_host`, have an `api_key` attribute,
    but existing hosts retrieved with `conjur.API.host` or the constructor of this class *do not*
    have one.

    Example:

        >>> # Create a host and save it's api key to a file.
        >>> host = api.create_host('jenkins')
        >>> api_key = host.api_key
        >>> with open('/etc/conjur.identity') as f:
        ...     f.write(api_key)

    Example:

        >>> # See if a host named `jenkins` exists:
        >>> if api.host('jenkins').exists():
        ...     print("Host 'jenkins' exists")
        ... else:
        ...     print("Host 'jenkins' does not exist")

    """
    def __init__(self, api, id, attrs=None):
        self.api = api
        self.id = id
        self._attrs = attrs
        self.role = self.api.role('host', self.id)

    def exists(self):
        """
        Return `True` if this host exists.
        """
        status = self.api.get(self._url(), check_errors=False).status_code
        if status == 200:
            return True
        if status == 404:
            return False
        raise ConjurException("Request Failed: {0}".format(status))

    def _fetch(self):
        self._attrs = self.api.get(self._url()).json()

    def _url(self):
        return "{0}/hosts/{1}".format(self.api.config.core_url,
                                      urlescape(self.id))

    def __getattr__(self, item):
        if self._attrs is None:
            self._fetch()
        try:
            return self._attrs[item]
        except KeyError:
            raise AttributeError(item)



#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from mock import patch, Mock
import requests

import conjur


@patch.object(requests, 'post')
def test_create_variable(mock_post):
    mock_post.return_value = mock_response = Mock()
    mock_response.status_code = 201
    mock_response.json = lambda: {'id': 'foobar'}  # No attribute support now
    api = conjur.new_from_token('token')
    v = api.create_variable(mime_type='mimey', kind='something')
    assert v.id == 'foobar'
    mock_post.assert_called_with(
        '%s/variables' % api.config.core_url,
        data={'mime_type': 'mimey', 'kind': 'something'},
        headers={'Authorization': api.auth_header()},
        verify=api.config.verify
    )


@patch.object(requests, 'get')
def test_get_variable_value(mock_get):
    mock_get.return_value = resp = Mock()
    resp.status_code = 200
    resp.text = 'teh value'
    api = conjur.new_from_token('token')
    v = api.variable('my-id')
    assert v.value() == 'teh value'
    mock_get.assert_called_with(
        '%s/variables/my-id/value' % api.config.core_url,
        headers={'Authorization': api.auth_header()},
        verify=api.config.verify
    )


@patch.object(requests, 'post')
def test_add_variable_value(mock_post):
    mock_post.return_value = resp = Mock()
    resp.status_code = 201
    api = conjur.new_from_token('token')
    v = api.variable('var')
    v.add_value('boo')
    mock_post.assert_called_with(
        '%s/variables/var/values' % api.config.core_url,
        headers={'Authorization': api.auth_header()},
        data={'value': 'boo'},
        verify=api.config.verify
    )



#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from mock import patch, Mock
import requests

import conjur


@patch.object(requests, 'post')
def test_create_user(mock_post):
    api = conjur.new_from_token('token')
    mock_post.return_value = resp = Mock()
    resp.status_code = 200
    resp.json = lambda:  {'login': 'foo', 'api_key': 'apikey'}

    user_no_pass = api.create_user('foo')
    assert user_no_pass.login == 'foo'
    assert user_no_pass.api_key == 'apikey'
    mock_post.assert_called_with(
        '{0}/users'.format(api.config.core_url),
        data={'login': 'foo'},
        headers={'Authorization': api.auth_header()},
        verify=api.config.verify
    )

    api.create_user('foo', 'bar')
    mock_post.assert_called_with(
        '{0}/users'.format(api.config.core_url),
        data={'login': 'foo', 'password': 'bar'},
        headers={'Authorization': api.auth_header()},
        verify=api.config.verify
    )


@patch.object(requests, 'get')
def test_user(mock_get):
    api = conjur.new_from_token('token')
    mock_get.return_value = Mock(status_code=200, json=lambda: {'foo': 'bar'})
    user = api.user('login')
    assert user.foo == 'bar'
    mock_get.assert_called_with(
        '{0}/users/login'.format(api.config.core_url),
        headers={'Authorization': api.auth_header()},
        verify=api.config.verify
    )


def test_user_role():
    user = conjur.new_from_key('foo', 'bar').user('someone')
    role = user.role
    assert role.kind == 'user'
    assert role.identifier == 'someone'



#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from mock import patch, Mock, call
import conjur

api = conjur.new_from_key('fakeid', 'fakepass')


@patch.object(api, 'get')
def test_public_keys(mock_get):
    response = "a b key1\na b key2"
    mock_get.return_value = Mock(text=response)
    assert api.public_keys('foo bar') == response
    mock_get.assert_called_with(
        '{0}/foo%20bar'.format(api.config.pubkeys_url)
    )


@patch.object(api, 'get')
def test_public_key(mock_get):
    mock_get.return_value = Mock(text="a b c")
    assert api.public_key('foo bar', 'keyname') == 'a b c'
    mock_get.assert_called_with(
        '{0}/foo%20bar/keyname'.format(api.config.pubkeys_url))


@patch.object(api, 'get')
def test_public_key_names(mock_get):
    response = "a b key1\na b key2"
    mock_get.return_value = Mock(text=response)
    assert list(api.public_key_names('foo bar')) == ['key1', 'key2']
    mock_get.assert_called_with(
        '{0}/foo%20bar'.format(api.config.pubkeys_url)
    )


@patch.object(api, 'post')
def test_add_public_key(mock_post):
    api.add_public_key('foo', 'a b c')
    mock_post.assert_called_with(api.config.pubkeys_url + '/foo', data='a b c')


@patch.object(api, 'delete')
def test_remove_public_key(mock_del):
    api.remove_public_key('foo', 'bar')
    mock_del.assert_called_with(api.config.pubkeys_url + '/foo/bar')


@patch.object(api, 'delete')
@patch.object(api, 'get')
def test_remove_public_keys(mock_get, mock_del):
    mock_get.return_value = Mock(text="a b key1\na b key2")
    api.remove_public_keys('foo')
    mock_del.assert_has_calls([
        call(api.config.pubkeys_url + '/foo/key1'),
        call(api.config.pubkeys_url + '/foo/key2')
        ])



#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import base64

import requests

from conjur.variable import Variable
from conjur.user import User
from conjur.role import Role
from conjur.group import Group
from conjur.layer import Layer
from conjur.host import Host
from conjur.resource import Resource
from conjur.util import urlescape
from conjur.exceptions import ConjurException

class API(object):
    def __init__(self, credentials=None, token=None, config=None):
        """
        Creates an API instance configured with the given credentials or token
        and config.

        Generally you should use `conjur.new_from_key`, `conjur.new_from_netrc`,
        or `conjur.new_from_token` to get an API instance instead of calling
        this constructor directly.

        `credentials` should be a `(login,api_key)` tuple if present.

        `token` should be a string containing a Conjur JSON token to use when authenticating.

        `config` should be a `conjur.Config` instance, and defaults to `conjur.config`.
        """
        if credentials:
            self.login, self.api_key = credentials
            self.token = None
        elif token:
            self.token = token
            self.login = self.api_key = None
        else:
            raise TypeError("must be given a credentials or token argument")
        if config:
            self.config = config
        else:
            from conjur.config import config as default_config

            self.config = default_config

    def authenticate(self, cached=True):
        """
        Authenticate with Conjur and return a token (str) that can be used
        to establish identity to Conjur services.

        Returns the json formatted signed Conjur authentication Token.

        It is an error to call this method if the API was created with a token
        rather than a login and api key.

        When `cached` is True, a cached token value will be used if it is
        available, otherwise the token will be fetched whether or not a
        cached value is present.
        """
        if cached and self.token:
            return self.token

        if not self.login or not self.api_key:
            raise ConjurException(
                "API created without credentials can't authenticate")

        url = "%s/users/%s/authenticate" % (self.config.authn_url,
                                            urlescape(self.login))

        self.token = self._request('post', url, self.api_key).text
        return self.token

    def auth_header(self):
        """
        Get the value of an Authorization header to make Conjur requests,
        performing authentication if necessary.

        Returns a string suitable for use as an `Authorization` header value.
        """
        token = self.authenticate()
        enc = base64.b64encode(token)
        return 'Token token="%s"' % enc

    def request(self, method, url, **kwargs):
        """
        Make an authenticated request with the given method and url.
        Additional arguments are passed to requests.<method>.

        Returns a `requests.Response` object.

        If the response status is not 2xx, raises a `conjur.ConjurException`.

        `method` is of the standard HTTP verbs (case insensitive).

        `url` is the full url to request.

        If `kwargs['check_errors']` is `True`, non-2xx responses will raise a `conjur.ConjurException`.
        Otherwise, it is the callers responsibility to check the status of the returned `requests.Response`.

        Additional are passed through to the requests.<method> call after adding an `Authorization`
        header and HTTPS verification settings.
        """
        headers = kwargs.setdefault('headers', {})
        headers['Authorization'] = self.auth_header()

        return self._request(method, url, **kwargs)

    def _request(self, method, url, *args, **kwargs):
        if 'verify' not in kwargs:
            kwargs['verify'] = self.config.verify
        check_errors = kwargs.pop('check_errors', True)

        response = getattr(requests, method.lower())(url, *args, **kwargs)
        if check_errors and response.status_code >= 300:
            raise ConjurException("Request failed: %d" % response.status_code)

        return response

    def get(self, url, **kwargs):
        """
        **NOTE** You will generally not need to use this method directly.

        Makes an authenticated GET request to the given :url:.

        Returns a requests.Response object.

        If the response status is not 2xx, raises a ConjurException.

        `url` is the full url to request.
         Keyword arguments are passed through to `requests.get`.
        """
        return self.request('get', url, **kwargs)

    def post(self, url, **kwargs):
        """
        **NOTE** You will generally not need to use this method directly.

        Makes an authenticated `POST` request to the given `url`

        Returns a `requests.Response` object.

        If the response status is not 2xx, raises a `conjur.ConjurException`.

        `url` is the full url to request.
         Keyword arguments are passed through to `requests.post`.
        """
        return self.request('post', url, **kwargs)

    def put(self,url, **kwargs):
        """
        **NOTE** You will generally not need to use this method directly.

        Makes an authenticated `PUT` request to the given `url`

        Returns a `requests.Response` object.

        If the response status is not 2xx, raises a `conjur.ConjurException`.

        `url` is the full url to request.
         Keyword arguments are passed through to `requests.put`.
        """
        return self.request('put', url, **kwargs)

    def delete(self, url, **kwargs):
        """
        **NOTE** You will generally not need to use this method directly.

        Makes an authenticated `DELETE` request to the given `url`

        Returns a `requests.Response` object.

        If the response status is not 2xx, raises a `conjur.ConjurException`.

        `url` is the full url to request.
         Keyword arguments are passed through to `requests.delete`.
        """
        return self.request('delete', url, **kwargs)

    def role(self, kind, identifier):
        """
        Return a `conjur.Role` object with the given kind and id.

        This method neither creates nor checks for the roles's existence.

        `kind` should be the role kind - for example, `"group"`, `"user"`,
        or `"host"`.

        `identifier` should be the *unqualified* Conjur id.  For example, to
        get the role for a user named bub, you would call `api.role('user', 'bub')`.
        """
        return Role(self, kind, identifier)

    def resource(self, kind, identifier):
        """
        Return a `conjur.Resource` object with the given kind and id.

        This method neither creates nor checks for the resource's existence.

        `kind` should be the resource kind - for example, `"variable"`, `"webservice"`,
        or `"configuration"`.

        `identifier` should be the *unqualified* Conjur id.  For example, to
        get the resource for a user variable named db_password, you would call
        `api.resource('variable', 'db_password')`.
        """
        return Resource(self, kind, identifier)

    def group(self, id):
        """
        Return a `conjur.Group` object with the given id.

        This method neither creates nor checks for the groups's existence.

        `id` is the *unqualified* id of the group, and does not include the account or kind.
        """
        return Group(self, id)

    def create_group(self, id):
        """
        Creates a Conjur Group and returns a `conjur.Group` object representing it.

        `id` is the identifier of the group to create.
        """

        self.post('{0}/groups'.format(self.config.core_url), data={'id': id})
        return Group(self, id)

    def variable(self, id):
        """
        Return a `conjur.Variable` object with the given `id`.

        This method neither creates nor checks for the variable's existence.
        """
        return Variable(self, id)

    def create_variable(self, id=None, mime_type='text/plain', kind='secret',
                        value=None):
        """
        Creates a Conjur variable.

        Returns a `conjur.Variable` object.

        `id` is an identifier for the new variable.  If not given, a unique id will
        be generated.

        `mime_type` is a string like `text/plain` indicating the content type stored by the
        variable.  This determines the Content-Type header of responses returning the variable's value.

        `kind`  is a string indicating a user defined role for the variable.
        Ignored by Conjur, but useful for making a variable's
        purpose.

        `value` is a string assigning an initial value for the variable.
        """
        data = {'mime_type': mime_type, 'kind': kind}
        if id is not None:
            data['id'] = id
        if value is not None:
            data['value'] = value

        attrs = self.post("%s/variables" % self.config.core_url, data=data).json()
        id = id or attrs['id']
        return Variable(self, id, attrs)

    def layer(self, layer_id):
        """
        Return a `conjur.Layer` object with the given `layer_id`.

        This method neither creates nor checks for the layer's existence.
        """
        return Layer(self, layer_id)

    def host(self, host_id):
        """
        Return a `conjur.Host` object with the given `host_id`.

        This method neither creates nor checks for the host's existence.
        """
        return Host(self, host_id)

    def create_host(self, host_id):
        """
        Creates a Conjur Host and returns a `conjur.Host` object that represents it.

        `host_id` is the id of the Host to be created.  The `conjur.Host` object returned by
        this method will have an `api_key` attribute, but when the Host is fetched in the future this attribute
        is not available.
        """
        attrs = self.post("{0}/hosts".format(self.config.core_url),
                          data={'id': host_id}).json()
        return Host(self, host_id, attrs)

    def user(self, login):
        """
        Returns an object representing a Conjur user with the given login.

        The user is *not* created by this method, and may in fact not exist.
        """
        return User(self, login)

    def create_user(self, login, password=None):
        """
        Create a Conjur user with the given `login` and password, and returns a `conjur.User` object
        representing it.

        If `password` is not given, the user will only be able to authenticate using the generated api_key
        attribute of the returned User instance.  Note that this `api_key` will not be available when the User
        is fetched in the future.
        """
        data = {'login': login}
        if password is not None:
            data['password'] = password
        url = "{0}/users".format(self.config.core_url)
        return User(self, login, self.post(url, data=data).json())

    def _public_key_url(self, *args):
        return '/'.join([self.config.pubkeys_url] +
                        [urlescape(arg) for arg in args])

    def add_public_key(self, username, key):
        """
        Upload an openssh formatted public key to be made available for the user
        given by `username`.

        The key should be formatted like `ssh-rsa <data...> bob@example.com`.
        """
        self.post(self._public_key_url(username), data=key)

    def remove_public_key(self, username, keyname):
        """
        Remove a specific public key for the user identified by `username`.
        The `keyname` argument refers to the name field in the openssh formatted key
        to be deleted.

        For example, if they key contents are `ssh-rsa <data...> bob@example.com`,
        the `keyname` should be `bob@example.com`
        """
        self.delete(self._public_key_url(username, keyname))

    def remove_public_keys(self, username):
        """
        Remove all public keys for the user represented by `username`.
        """
        for keyname in self.public_key_names(username):
            self.remove_public_key(username, keyname)

    def public_keys(self, username):
        """
        Returns all keys for the user given by `username`, as a newline delimited string.

        The odd format is chosen to support the Conjur SSH login implementation.
        """
        return self.get(self._public_key_url(username)).text

    def public_key(self, username, keyname):
        """
        Return the contents of a specific public key given by `keyname`,
        for the user given by `username` as a string.

        The name of the key is based on the name entry of the openssh formatted key that was uploaded.

        For example, if they key contents are `ssh-rsa <data...> bob@example.com`,
        the `keyname` should be `bob@example.com`
        """
        return self.get(self._public_key_url(username, keyname)).text

    def public_key_names(self, username):
        """
        Return the names of public keys for the user given by `username`.

        The names of the keys are based on the name entry of the openssh formatted key that was uploaded.

        For example, if they key contents are `ssh-rsa <data...> bob@example.com`,
        the `keyname` should be `bob@example.com`
        """
        return [k.split(' ')[-1] for k in self.public_keys(username).split('\n')]




#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# 2 vs 3 urlencode moved...
try:
    from urllib import quote
except:
    from urllib.parse import quote


def urlescape(s):
    return quote(s, '')


def authzid(obj, kind, with_account=True):
    if isinstance(obj, (str, unicode)):  # noqa F821 (flake8 doesn't know about unicode)
        if not with_account:
            return ':'.join(obj.split(':')[1:])
        return obj
    for attr in (kind, kind + 'id'):
        if hasattr(obj, attr):
            return authzid(getattr(obj, attr), kind)
    raise TypeError("Can't get {0}id from {1}".format(kind, obj))



#
# Copyright (C) 2016 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from mock import patch, Mock
import conjur
import pytest

api = conjur.new_from_key('admin', 'secret')
resource = api.resource('food', 'bacon')
bob = api.user('bob')


def test_resource_id():
    assert resource.resourceid == api.config.account + ':food:bacon'


@patch.object(api, 'get')
def test_permitted_with_role(mock_get):
    mock_get.return_value = Mock(status_code=204)
    assert resource.permitted('fry', bob)

    mock_get.assert_called_with(
        'https://example.com/api/authz/conjur/roles/user/bob',
        params={'privilege': 'fry', 'check': 'true',
                'resource_id': 'conjur:food:bacon'},
        check_errors=False
    )


@patch.object(api, 'get')
def test_permitted_self_role(mock_get):
    mock_get.return_value = Mock(status_code=204)
    assert resource.permitted('fry')
    mock_get.assert_called_with(
        'https://example.com/api/authz/conjur/resources/food/bacon',  # noqa E501 (line too long)
        params={'privilege': 'fry', 'check': 'true'},
        check_errors=False
    )


@patch.object(api, 'get')
def test_permitted_fails_with_self_role(mock_get):
    mock_get.return_value = Mock(status_code=403)
    assert not resource.permitted('fry')


@patch.object(api, 'get')
def test_permitted_fails_with_role(mock_get):
    mock_get.return_value = Mock(status_code=403)
    assert not resource.permitted('fry', bob)


@patch.object(api, 'get')
def test_permitted_error_self_role(mock_get):
    mock_get.return_value = Mock(status_code=401)
    with pytest.raises(conjur.ConjurException):
        resource.permitted('fry')


@patch.object(api, 'get')
def test_permitted_error_with_role(mock_get):
    mock_get.return_value = Mock(status_code=401)
    with pytest.raises(conjur.ConjurException):
        resource.permitted('fry', bob)



from setuptools import setup, find_packages

NAME = 'Conjur'
VERSION = '0.4.4'

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    print("Warning: Unable to convert README.md to rst!")
    long_description = open('README.md').read()


setup(
    name=NAME,
    version=VERSION,
    packages=find_packages(),
    install_requires=open('requirements.txt').readlines(),
    package_data={},
    author='Jon Mason',
    author_email='jon@conjur.net',
    description='Python client for the Conjur API',
    long_description=long_description,
    license='MIT',
    url='https://github.com/conjurinc/api-python',
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7'
    ]
)



#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from conjur.util import urlescape


class Variable(object):
    """
    A `Variable` represents a versioned secret stored in Conjur.

    Generally you will get an instance of this class by calling `conjur.API.create_variable`
    or `conjur.API.variable`.

    Instances of this class allow you to fetch values of the variable, and store new ones.

    Example:

        >>> # Print the current value of the variable `mysql-password`
        >>> variable = api.variable('mysql-password')
        >>> print("mysql-password is {}".format(variable.value()))

    Example:

        >>> # Print all versions of the same variable
        >>> variable = api.variable('mysql-password')
        >>> for i in range(1, variable.version_count + 1): # version numbers are 1 based
        ...     print("version {} of 'mysql-password' is {}".format(i, variable.value(i)))

    """
    def __init__(self, api, id, attrs=None):
        self.id = id
        self.api = api
        self._attrs = attrs

    def value(self, version=None):
        """
        Retrieve the secret stored in a variable.

        `version` is a *one based* index of the version to be retrieved.

        If no such version exists, a 404 error is raised.

        Returns the value of the variable as a string.
        """
        url = "%s/variables/%s/value" % (self.api.config.core_url,
                                         urlescape(self.id))
        if version is not None:
            url = "%s?version=%s" % (url, version)
        return self.api.get(url).text

    def add_value(self, value):
        """
        Stores a new version of the secret in this variable.

        `value` is a string giving the new value to store.

        This increments the variable's `version_count` member by one.
        """
        self._attrs = None
        data = {'value': value}
        url = "%s/variables/%s/values" % (self.api.config.core_url,
                                          urlescape(self.id))
        self.api.post(url, data=data)

    def __getattr__(self, item):
        if self._attrs is None:
            self._fetch()
        try:
            return self._attrs[item]
        except KeyError:
            raise AttributeError(item)

    def _fetch(self):
        self._attrs = self.api.get(
            "{0}/variables/{1}".format(self.api.config.core_url,
                                       urlescape(self.id))
        ).json()



#
# Copyright (C) 2014 Conjur Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from conjur.util import urlescape, authzid
from conjur.exceptions import ConjurException
import logging


class Role(object):
    """
    Represents a Conjur [role](https://developer.conjur.net/key_concepts/rbac.html#rbac-roles).

    An instance of this class does not know whether the role in question exists.

    Generally you should create instances of this class through the `conjur.API.role` method,
    or the `Role.from_roleid` classmethod.

    Roles can provide information about their members and can check whether the role they represent
    is allowed to perform certain operations on resources.

    `conjur.User` and `conjur.Group` objects have `role` members that reference the role corresponding
    to that Conjur asset.
    """
    def __init__(self, api, kind, identifier):
        """
        Create a role to represent the Conjur role with id `<kind>:<identifier>`.  For
        example, to represent the role associated with a user named bob,

            role = Role(api, 'user', 'bob')

        `api` must be a `conjur.API` instance, used to implement this classes interactions with Conjur

        `kind` is a string giving the role kind

        `identifier` is the unqualified identifier of the role.
        """

        self.api = api
        """
        The `conjur.API` instance used to implement our methods.
        """

        self.kind = kind
        """
        The `kind` portion of the role's id.
        """

        self.identifier = identifier
        """
        The `identifier` portion of the role's id.
        """

    @classmethod
    def from_roleid(cls, api, roleid):
        """
        Creates an instance of `conjur.Role` from a full role id string.

        `api` is an instance of `conjur.API`

        `roleid` is a fully or partially qualified Conjur identifier, for example,
        `"the-account:service:some-service"` or `"service:some-service"` resolve to the same role.
        """
        tokens = authzid(roleid, 'role').split(':', 3)
        if len(tokens) == 3:
            tokens.pop(0)
        return cls(api, *tokens)

    @property
    def roleid(self):
        """
        Return the full role id as a string.

        Example:

            >>> role = api.role('user', 'bob')
            >>> role.roleid
            'the-account:user:bob'

         """
        return ':'.join([self.api.config.account, self.kind, self.identifier])

    def is_permitted(self, resource, privilege):
        """
        Check whether `resource` has `privilege` on this role.

        `resource` is a qualified identifier for the resource.

        `privilege` is a string like `"update"` or `"execute"`.


        Example:

            >>> role = api.role('user', 'alice')
            >>> if role.is_permitted('variable:db-password', 'execute'):
            ...     print("Alice can fetch the database password")
            ... else:
            ...     print("Alice cannot fetch the database password")

        """
        params = {
            'check': 'true',
            'resource_id': authzid(resource, 'resource'),
            'privilege': privilege
        }
        response = self.api.get(self._url(), params=params,
                                check_errors=False)
        if response.status_code == 204:
            return True
        elif response.status_code in (403, 404):
            return False
        else:
            raise ConjurException("Request failed: %d" % response.status_code)

    def grant_to(self, member, admin=None):
        """
        Grant this role to `member`.

        `member` is a string or object with a `role` attribute or `roleid` method,
            such as a `conjur.User` or `conjur.Group`.

        `admin` whether the member can grant this role to others.

        """
        data = {}
        if admin is not None:
            data['admin'] = 'true' if admin else 'false'
        self.api.put(self._membership_url(member), data=data)

    def revoke_from(self, member):
        """
        The inverse of `conjur.Role.grant_to`.  Removes `member` from the members of this
        role.

        `member` is a string or object with a `role` attribute or `roleid` method,
            such as a `conjur.User` or `conjur.Group`.
        """
        self.api.delete(self._membership_url(member))

    def members(self):
        """
        Return a list of members of this role.  Members are returned as `dict`s
        with the following keys:

        * `'member'` the fully qualified identifier of the group
        * `'role'` the fully qualified identifier of the group (redundant)
        * `'grantor'` the role that granted the membership
        * `'admin_option'` whether this member can grant membership in the group to other roles.
        """
        return self.api.get(self._membership_url()).json()

    def _membership_url(self, member=None):
        url = self._url() + "?members"
        if member is not None:
            memberid = authzid(member, 'role')
            url += "&member=" + urlescape(memberid)
        return url

    def _url(self, *args):
        return "/".join([self.api.config.authz_url,
                         self.api.config.account,
                         'roles',
                         self.kind,
                         self.identifier] + list(args))
